@@ -5,14 +5,14 @@ from typing import Any, Callable, Dict, List
 
 import isort.comments
 
-_wrap_modes: Dict[str, Callable[[Any], str]] = {}
+_wrap_modes: Dict[str, Callable[..., str]] = {}
 
 
 def from_string(value: str) -> "WrapModes":
     return getattr(WrapModes, str(value), None) or WrapModes(int(value))
 
 
-def formatter_from_string(name: str):
+def formatter_from_string(name: str) -> Callable[..., str]:
     return _wrap_modes.get(name.upper(), grid)
 
 
@@ -32,18 +32,18 @@ def _wrap_mode_interface(
     return ""
 
 
-def _wrap_mode(function):
+def _wrap_mode(function: Callable[..., str]) -> Callable[..., str]:
     """Registers an individual wrap mode. Function name and order are significant and used for
     creating enum.
     """
     _wrap_modes[function.__name__.upper()] = function
-    function.__signature__ = signature(_wrap_mode_interface)
+    function.__signature__ = signature(_wrap_mode_interface)  # type: ignore
     function.__annotations__ = _wrap_mode_interface.__annotations__
     return function
 
 
 @_wrap_mode
-def grid(**interface):
+def grid(**interface: Any) -> str:
     if not interface["imports"]:
         return ""
 
@@ -80,11 +80,11 @@ def grid(**interface):
             interface["comments"] = []
         else:
             interface["statement"] += ", " + next_import
-    return interface["statement"] + ("," if interface["include_trailing_comma"] else "") + ")"
+    return f"{interface['statement']}{',' if interface['include_trailing_comma'] else ''})"
 
 
 @_wrap_mode
-def vertical(**interface):
+def vertical(**interface: Any) -> str:
     if not interface["imports"]:
         return ""
 
@@ -106,68 +106,70 @@ def vertical(**interface):
     return f"{interface['statement']}({first_import}{_imports}{_comma_maybe})"
 
 
-def _hanging_indent_common(use_parentheses=False, **interface):
+def _hanging_indent_end_line(line: str) -> str:
+    if not line.endswith(" "):
+        line += " "
+    return line + "\\"
+
+
+@_wrap_mode
+def hanging_indent(**interface: Any) -> str:
     if not interface["imports"]:
         return ""
-    line_length_limit = interface["line_length"] - (1 if use_parentheses else 3)
 
-    def end_line(line):
-        if use_parentheses:
-            return line
-        if not line.endswith(" "):
-            line += " "
-        return line + "\\"
+    line_length_limit = interface["line_length"] - 3
 
-    if use_parentheses:
-        interface["statement"] += "("
     next_import = interface["imports"].pop(0)
     next_statement = interface["statement"] + next_import
     # Check for first import
     if len(next_statement) > line_length_limit:
         next_statement = (
-            isort.comments.add_to_line(
-                interface["comments"],
-                end_line(interface["statement"]),
-                removed=interface["remove_comments"],
-                comment_prefix=interface["comment_prefix"],
-            )
-            + f"{interface['line_separator']}{interface['indent']}{next_import}"
+            _hanging_indent_end_line(interface["statement"])
+            + interface["line_separator"]
+            + interface["indent"]
+            + next_import
         )
-        interface["comments"] = []
+
     interface["statement"] = next_statement
     while interface["imports"]:
         next_import = interface["imports"].pop(0)
-        next_statement = isort.comments.add_to_line(
+        next_statement = interface["statement"] + ", " + next_import
+        if len(next_statement.split(interface["line_separator"])[-1]) > line_length_limit:
+            next_statement = (
+                _hanging_indent_end_line(interface["statement"] + ",")
+                + f"{interface['line_separator']}{interface['indent']}{next_import}"
+            )
+        interface["statement"] = next_statement
+
+    interface[
+        "statement"
+    ] = f"{interface['statement']}{',' if interface['include_trailing_comma'] else ''}"
+    if interface["comments"]:
+        statement_with_comments = isort.comments.add_to_line(
             interface["comments"],
-            interface["statement"] + ", " + next_import,
+            interface["statement"],
             removed=interface["remove_comments"],
             comment_prefix=interface["comment_prefix"],
         )
-        current_line = next_statement.split(interface["line_separator"])[-1]
-        if len(current_line) > line_length_limit:
-            next_statement = (
-                isort.comments.add_to_line(
-                    interface["comments"],
-                    end_line(interface["statement"] + ","),
-                    removed=interface["remove_comments"],
-                    comment_prefix=interface["comment_prefix"],
-                )
-                + f"{interface['line_separator']}{interface['indent']}{next_import}"
+        if len(statement_with_comments.split(interface["line_separator"])[-1]) <= (
+            line_length_limit + 2
+        ):
+            return statement_with_comments
+        return (
+            _hanging_indent_end_line(interface["statement"])
+            + str(interface["line_separator"])
+            + isort.comments.add_to_line(
+                interface["comments"],
+                interface["indent"],
+                removed=interface["remove_comments"],
+                comment_prefix=interface["comment_prefix"].lstrip(),
             )
-            interface["comments"] = []
-        interface["statement"] = next_statement
-    _comma_maybe = "," if interface["include_trailing_comma"] else ""
-    _close_parentheses_maybe = ")" if use_parentheses else ""
-    return interface["statement"] + _comma_maybe + _close_parentheses_maybe
+        )
+    return str(interface["statement"])
 
 
 @_wrap_mode
-def hanging_indent(**interface):
-    return _hanging_indent_common(use_parentheses=False, **interface)
-
-
-@_wrap_mode
-def vertical_hanging_indent(**interface):
+def vertical_hanging_indent(**interface: Any) -> str:
     _line_with_comments = isort.comments.add_to_line(
         interface["comments"],
         "",
@@ -182,7 +184,7 @@ def vertical_hanging_indent(**interface):
     )
 
 
-def _vertical_grid_common(need_trailing_char: bool, **interface):
+def _vertical_grid_common(need_trailing_char: bool, **interface: Any) -> str:
     if not interface["imports"]:
         return ""
 
@@ -201,9 +203,11 @@ def _vertical_grid_common(need_trailing_char: bool, **interface):
         next_import = interface["imports"].pop(0)
         next_statement = f"{interface['statement']}, {next_import}"
         current_line_length = len(next_statement.split(interface["line_separator"])[-1])
-        if interface["imports"] or need_trailing_char:
-            # If we have more interface["imports"] we need to account for a comma after this import
-            # We might also need to account for a closing ) we're going to add.
+        if interface["imports"] or interface["include_trailing_comma"]:
+            # We need to account for a comma after this import.
+            current_line_length += 1
+        if not interface["imports"] and need_trailing_char:
+            # We need to account for a closing ) we're going to add.
             current_line_length += 1
         if current_line_length > interface["line_length"]:
             next_statement = (
@@ -213,34 +217,32 @@ def _vertical_grid_common(need_trailing_char: bool, **interface):
         interface["statement"] = next_statement
     if interface["include_trailing_comma"]:
         interface["statement"] += ","
-    return interface["statement"]
+    return str(interface["statement"])
 
 
 @_wrap_mode
-def vertical_grid(**interface) -> str:
+def vertical_grid(**interface: Any) -> str:
     return _vertical_grid_common(need_trailing_char=True, **interface) + ")"
 
 
 @_wrap_mode
-def vertical_grid_grouped(**interface):
-    return (
-        _vertical_grid_common(need_trailing_char=True, **interface)
-        + interface["line_separator"]
-        + ")"
-    )
-
-
-@_wrap_mode
-def vertical_grid_grouped_no_comma(**interface):
+def vertical_grid_grouped(**interface: Any) -> str:
     return (
         _vertical_grid_common(need_trailing_char=False, **interface)
-        + interface["line_separator"]
+        + str(interface["line_separator"])
         + ")"
     )
 
 
 @_wrap_mode
-def noqa(**interface):
+def vertical_grid_grouped_no_comma(**interface: Any) -> str:
+    # This is a deprecated alias for vertical_grid_grouped above. This function
+    # needs to exist for backwards compatibility but should never get called.
+    raise NotImplementedError
+
+
+@_wrap_mode
+def noqa(**interface: Any) -> str:
     _imports = ", ".join(interface["imports"])
     retval = f"{interface['statement']}{_imports}"
     comment_str = " ".join(interface["comments"])
@@ -260,7 +262,7 @@ def noqa(**interface):
 
 
 @_wrap_mode
-def vertical_hanging_indent_bracket(**interface):
+def vertical_hanging_indent_bracket(**interface: Any) -> str:
     if not interface["imports"]:
         return ""
     statement = vertical_hanging_indent(**interface)
@@ -268,7 +270,7 @@ def vertical_hanging_indent_bracket(**interface):
 
 
 @_wrap_mode
-def vertical_prefix_from_module_import(**interface):
+def vertical_prefix_from_module_import(**interface: Any) -> str:
     if not interface["imports"]:
         return ""
 
@@ -304,18 +306,69 @@ def vertical_prefix_from_module_import(**interface):
 
     if comments and statement_with_comments:
         output_statement = statement_with_comments
-    return output_statement
+    return str(output_statement)
 
 
 @_wrap_mode
-def hanging_indent_with_parentheses(**interface):
-    return _hanging_indent_common(use_parentheses=True, **interface)
+def hanging_indent_with_parentheses(**interface: Any) -> str:
+    if not interface["imports"]:
+        return ""
+
+    line_length_limit = interface["line_length"] - 1
+
+    interface["statement"] += "("
+    next_import = interface["imports"].pop(0)
+    next_statement = interface["statement"] + next_import
+    # Check for first import
+    if len(next_statement) > line_length_limit:
+        next_statement = (
+            isort.comments.add_to_line(
+                interface["comments"],
+                interface["statement"],
+                removed=interface["remove_comments"],
+                comment_prefix=interface["comment_prefix"],
+            )
+            + f"{interface['line_separator']}{interface['indent']}{next_import}"
+        )
+        interface["comments"] = []
+    interface["statement"] = next_statement
+    while interface["imports"]:
+        next_import = interface["imports"].pop(0)
+        if (
+            not interface["line_separator"] in interface["statement"]
+            and "#" in interface["statement"]
+        ):
+            line, comments = interface["statement"].split("#", 1)
+            next_statement = (
+                f"{line.rstrip()}, {next_import}{interface['comment_prefix']}{comments}"
+            )
+        else:
+            next_statement = isort.comments.add_to_line(
+                interface["comments"],
+                interface["statement"] + ", " + next_import,
+                removed=interface["remove_comments"],
+                comment_prefix=interface["comment_prefix"],
+            )
+        current_line = next_statement.split(interface["line_separator"])[-1]
+        if len(current_line) > line_length_limit:
+            next_statement = (
+                isort.comments.add_to_line(
+                    interface["comments"],
+                    interface["statement"] + ",",
+                    removed=interface["remove_comments"],
+                    comment_prefix=interface["comment_prefix"],
+                )
+                + f"{interface['line_separator']}{interface['indent']}{next_import}"
+            )
+            interface["comments"] = []
+        interface["statement"] = next_statement
+    return f"{interface['statement']}{',' if interface['include_trailing_comma'] else ''})"
 
 
 @_wrap_mode
-def backslash_grid(**interface):
+def backslash_grid(**interface: Any) -> str:
     interface["indent"] = interface["white_space"][:-1]
-    return _hanging_indent_common(use_parentheses=False, **interface)
+    return hanging_indent(**interface)
 
 
 WrapModes = enum.Enum(  # type: ignore
