@@ -1,8 +1,9 @@
 import json
+import os
 import subprocess
 from datetime import datetime
-from io import BytesIO, TextIOWrapper
 
+import py
 import pytest
 from hypothesis import given
 from hypothesis import strategies as st
@@ -11,16 +12,14 @@ from isort import main
 from isort._version import __version__
 from isort.exceptions import InvalidSettingsPath
 from isort.settings import DEFAULT_CONFIG, Config
-from isort.wrap_modes import WrapModes
+from .utils import as_stream
+from io import BytesIO, TextIOWrapper
+from typing import TYPE_CHECKING, Any
 
-
-class UnseekableTextIOWrapper(TextIOWrapper):
-    def seek(self, *args, **kwargs):
-        raise ValueError("underlying stream is not seekable")
-
-
-def as_stream(text: str) -> UnseekableTextIOWrapper:
-    return UnseekableTextIOWrapper(BytesIO(text.encode("utf8")))
+if TYPE_CHECKING:
+    WrapModes: Any
+else:
+    from isort.wrap_modes import WrapModes
 
 
 @given(
@@ -43,15 +42,15 @@ def test_fuzz_sort_imports(file_name, config, check, ask_to_apply, write_to_stdo
 def test_sort_imports(tmpdir):
     tmp_file = tmpdir.join("file.py")
     tmp_file.write("import os, sys\n")
-    assert main.sort_imports(str(tmp_file), DEFAULT_CONFIG, check=True).incorrectly_sorted
+    assert main.sort_imports(str(tmp_file), DEFAULT_CONFIG, check=True).incorrectly_sorted  # type: ignore # noqa
     main.sort_imports(str(tmp_file), DEFAULT_CONFIG)
-    assert not main.sort_imports(str(tmp_file), DEFAULT_CONFIG, check=True).incorrectly_sorted
+    assert not main.sort_imports(str(tmp_file), DEFAULT_CONFIG, check=True).incorrectly_sorted  # type: ignore # noqa
 
     skip_config = Config(skip=["file.py"])
-    assert main.sort_imports(
+    assert main.sort_imports(  # type: ignore
         str(tmp_file), config=skip_config, check=True, disregard_skip=False
     ).skipped
-    assert main.sort_imports(str(tmp_file), config=skip_config, disregard_skip=False).skipped
+    assert main.sort_imports(str(tmp_file), config=skip_config, disregard_skip=False).skipped  # type: ignore # noqa
 
 
 def test_sort_imports_error_handling(tmpdir, mocker, capsys):
@@ -59,7 +58,7 @@ def test_sort_imports_error_handling(tmpdir, mocker, capsys):
     tmp_file.write("import os, sys\n")
     mocker.patch("isort.core.process").side_effect = IndexError("Example unhandled exception")
     with pytest.raises(IndexError):
-        main.sort_imports(str(tmp_file), DEFAULT_CONFIG, check=True).incorrectly_sorted
+        main.sort_imports(str(tmp_file), DEFAULT_CONFIG, check=True).incorrectly_sorted  # type: ignore # noqa
 
     out, error = capsys.readouterr()
     assert "Unrecoverable exception thrown when parsing" in error
@@ -78,6 +77,7 @@ def test_parse_args():
     assert main.parse_args(["--csi"]) == {"combine_straight_imports": True}
     assert main.parse_args(["--combine-straight-imports"]) == {"combine_straight_imports": True}
     assert main.parse_args(["--dont-follow-links"]) == {"follow_links": False}
+    assert main.parse_args(["--overwrite-in-place"]) == {"overwrite_in_place": True}
 
 
 def test_ascii_art(capsys):
@@ -145,6 +145,11 @@ sections=MADEUP
         main.main([str(python_file)])
 
 
+def test_ran_against_root():
+    with pytest.raises(SystemExit):
+        main.main(["/"])
+
+
 def test_main(capsys, tmpdir):
     base_args = [
         "-sp",
@@ -320,7 +325,7 @@ import a
 import b
 """
     )
-    main.main([str(tmpdir), "--skip", "skip.py", "--check"])
+    main.main([str(tmpdir), "--extend-skip", "skip.py", "--check"])
 
     # without filter options passed in should successfully sort files
     main.main([str(python_file), str(should_skip), "--verbose", "--atomic"])
@@ -334,20 +339,6 @@ import b
     out, error = capsys.readouterr()
     assert "Broken" in out
 
-    # should respect gitignore if requested.
-    out, error = capsys.readouterr()  # clear sysoutput before tests
-    subprocess.run(["git", "init", str(tmpdir)])
-    main.main([str(python_file), "--skip-gitignore", "--filter-files"])
-    out, error = capsys.readouterr()
-    assert "Skipped" not in out
-    tmpdir.join(".gitignore").write("has_imports.py")
-    main.main([str(python_file)])
-    out, error = capsys.readouterr()
-    assert "Skipped" not in out
-    main.main([str(python_file), "--skip-gitignore", "--filter-files"])
-    out, error = capsys.readouterr()
-    assert "Skipped" in out
-
     # warnings should be displayed if old flags are used
     with pytest.warns(UserWarning):
         main.main([str(python_file), "--recursive", "-fss"])
@@ -357,11 +348,6 @@ import b
         main.main(["-sp", str(config_file), "-"], stdin=input_content)
 
 
-def test_isort_command():
-    """Ensure ISortCommand got registered, otherwise setuptools error must have occurred"""
-    assert main.ISortCommand
-
-
 def test_isort_filename_overrides(tmpdir, capsys):
     """Tests isorts available approaches for overriding filename and extension based behavior"""
     input_text = """
@@ -892,6 +878,44 @@ __revision__ = 'יייי'
     out, error = capsys.readouterr()
 
 
+def test_stream_skip_file(tmpdir, capsys):
+    input_with_skip = """
+# isort: skip_file
+import b
+import a
+"""
+    stream_with_skip = as_stream(input_with_skip)
+    main.main(["-"], stdin=stream_with_skip)
+    out, error = capsys.readouterr()
+    assert out == input_with_skip
+
+    input_without_skip = input_with_skip.replace("isort: skip_file", "generic comment")
+    stream_without_skip = as_stream(input_without_skip)
+    main.main(["-"], stdin=stream_without_skip)
+    out, error = capsys.readouterr()
+    assert (
+        out
+        == """
+# generic comment
+import a
+import b
+"""
+    )
+
+    atomic_input_without_skip = input_with_skip.replace("isort: skip_file", "generic comment")
+    stream_without_skip = as_stream(atomic_input_without_skip)
+    main.main(["-", "--atomic"], stdin=stream_without_skip)
+    out, error = capsys.readouterr()
+    assert (
+        out
+        == """
+# generic comment
+import a
+import b
+"""
+    )
+
+
 def test_only_modified_flag(tmpdir, capsys):
     # ensures there is no verbose output for correct files with only-modified flag
 
@@ -1034,3 +1058,115 @@ def test_identify_imports_main(tmpdir, capsys):
     main.identify_imports_main(["-", "--attributes"], stdin=as_stream(file_content))
     out, error = capsys.readouterr()
     len(out.split("\n")) == 2
+
+
+def test_gitignore(capsys, tmpdir: py.path.local):
+
+    import_content = """
+import b
+import a
+"""
+
+    def main_check(args):
+        try:
+            main.main(args)
+        except SystemExit:
+            pass
+        return capsys.readouterr()
+
+    subprocess.run(["git", "init", str(tmpdir)])
+    python_file = tmpdir.join("has_imports.py")
+    python_file.write(import_content)
+    tmpdir.join("no_imports.py").write("...")
+
+    out, error = main_check([str(python_file), "--skip-gitignore", "--filter-files", "--check"])
+    assert "has_imports.py" in error and "no_imports.py" not in error
+
+    tmpdir.join(".gitignore").write("has_imports.py")
+
+    out, error = main_check([str(python_file), "--check"])
+    assert "has_imports.py" in error and "no_imports.py" not in error
+
+    out, error = main_check([str(python_file), "--skip-gitignore", "--filter-files", "--check"])
+    assert "Skipped" in out
+
+    # Should work with nested directories
+    tmpdir.mkdir("nested_dir")
+    tmpdir.join(".gitignore").write("nested_dir/has_imports.py")
+    subfolder_file = tmpdir.join("nested_dir/has_imports.py")
+    subfolder_file.write(import_content)
+
+    out, error = main_check([str(tmpdir), "--skip-gitignore", "--filter-files", "--check"])
+    assert "has_imports.py" in error and "nested_dir/has_imports.py" not in error
+
+    # Should work with relative path
+    currentdir = os.getcwd()
+    os.chdir(tmpdir)
+
+    out, error = main_check([".", "--skip-gitignore", "--filter-files", "--check"])
+    assert "has_imports.py" in error and "nested_dir/has_imports.py" not in error
+
+    tmpdir.join(".gitignore").write(
+        """
+nested_dir/has_imports.py
+has_imports.py
+"""
+    )
+    out, error = main_check([".", "--skip-gitignore", "--filter-files", "--check"])
+    assert "Skipped" in out
+
+    os.chdir(currentdir)
+
+    # Should work with multiple git projects
+
+    tmpdir.join(".git").remove()
+    tmpdir.join(".gitignore").remove()
+
+    # git_project0
+    # | has_imports_ignored.py ignored
+    # | has_imports.py should check
+    git_project0 = tmpdir.mkdir("git_project0")
+    subprocess.run(["git", "init", str(git_project0)])
+    git_project0.join(".gitignore").write("has_imports_ignored.py")
+    git_project0.join("has_imports_ignored.py").write(import_content)
+    git_project0.join("has_imports.py").write(import_content)
+
+    # git_project1
+    # | has_imports.py should check
+    # | nested_dir
+    #   | has_imports_ignored.py ignored
+    #   | has_imports.py should check
+    # | nested_dir_ignored ignored
+    #   | has_imports.py ignored from folder
+    git_project1 = tmpdir.mkdir("git_project1")
+    subprocess.run(["git", "init", str(git_project1)])
+    git_project1.join(".gitignore").write(
+        """
+nested_dir/has_imports_ignored.py
+nested_dir_ignored
+"""
+    )
+    git_project1.join("has_imports.py").write(import_content)
+    nested_dir = git_project1.mkdir("nested_dir")
+    nested_dir.join("has_imports.py").write(import_content)
+    nested_dir.join("has_imports_ignored.py").write(import_content)
+    git_project1.mkdir("nested_dir_ignored").join("has_imports.py").write(import_content)
+
+    should_check = [
+        "/has_imports.py",
+        "/nested_dir/has_imports.py",
+        "/git_project0/has_imports.py",
+        "/git_project1/has_imports.py",
+        "/git_project1/nested_dir/has_imports.py",
+    ]
+
+    out, error = main_check([str(tmpdir), "--skip-gitignore", "--filter-files", "--check"])
+
+    if os.name == "nt":
+        should_check = [sc.replace("/", "\\") for sc in should_check]
+
+    assert all(f"{str(tmpdir)}{file}" in error for file in should_check)
+
+    out, error = main_check([str(tmpdir), "--skip-gitignore", "--filter-files"])
+
+    assert all(f"{str(tmpdir)}{file}" in out for file in should_check)
