@@ -1,6 +1,8 @@
 """A growing set of tests designed to ensure isort doesn't have regressions in new versions"""
 from io import StringIO
 
+import pytest
+
 import isort
 
 
@@ -231,7 +233,9 @@ def test_ensure_sre_parse_is_identified_as_stdlib_issue_1304():
     """Ensure sre_parse is idenified as STDLIB.
     See: https://github.com/pycqa/isort/issues/1304.
     """
-    assert isort.place_module("sre_parse") == isort.place_module("sre") == isort.settings.STDLIB
+    assert (
+        isort.place_module("sre_parse") == isort.place_module("sre") == isort.settings.STDLIB  # type: ignore # noqa
+    )
 
 
 def test_add_imports_shouldnt_move_lower_comments_issue_1300():
@@ -716,6 +720,81 @@ import os
     )
 
 
+def test_isort_float_to_top_with_sort_on_off_tests():
+    """Characterization test for current behaviour of float-to-top on isort: on/off sections.
+    - imports in isort:off sections stay where they are
+    - imports in isort:on sections float up, but to the top of the isort:on section (not the
+      top of the file)"""
+    assert (
+        isort.code(
+            """
+def foo():
+    pass
+
+import a
+
+# isort: off
+import stays_in_section
+
+x = 1
+
+import stays_in_place
+
+# isort: on
+
+def bar():
+    pass
+
+import floats_to_top_of_section
+
+def baz():
+    pass
+""",
+            float_to_top=True,
+        )
+        == """import a
+
+
+def foo():
+    pass
+
+# isort: off
+import stays_in_section
+
+x = 1
+
+import stays_in_place
+
+# isort: on
+import floats_to_top_of_section
+
+
+def bar():
+    pass
+
+
+def baz():
+    pass
+"""
+    )
+
+    to_sort = """# isort: off
+
+def foo():
+    pass
+
+import stays_in_place
+import no_float_to_to_top
+import no_ordering
+
+def bar():
+    pass
+"""
+
+    # No changes if isort is off
+    assert isort.code(to_sort, float_to_top=True) == to_sort
+
+
 def test_isort_doesnt_float_to_top_correctly_when_imports_not_at_top_issue_1382():
     """isort should float existing imports to the top, if they are currently below the top.
     See: https://github.com/PyCQA/isort/issues/1382
@@ -1083,6 +1162,83 @@ def generator_function():
     assert isort.check_code(raise_from_at_file_end_ignored, show_diff=True)
 
 
+def test_isort_float_to_top_correctly_identifies_single_line_comments_1499():
+    """Test to ensure isort correctly handles the case where float to top is used
+    to push imports to the top and the top comment is a multiline type but only
+    one line.
+    See: https://github.com/PyCQA/isort/issues/1499
+    """
+    assert (
+        isort.code(
+            '''#!/bin/bash
+"""My comment"""
+def foo():
+    pass
+
+import a
+
+def bar():
+    pass
+''',
+            float_to_top=True,
+        )
+        == (
+            '''#!/bin/bash
+"""My comment"""
+import a
+
+
+def foo():
+    pass
+
+
+def bar():
+    pass
+'''
+        )
+    )
+    assert (
+        isort.code(
+            """#!/bin/bash
+'''My comment'''
+def foo():
+    pass
+
+import a
+
+def bar():
+    pass
+""",
+            float_to_top=True,
+        )
+        == (
+            """#!/bin/bash
+'''My comment'''
+import a
+
+
+def foo():
+    pass
+
+
+def bar():
+    pass
+"""
+        )
+    )
+
+    assert isort.check_code(
+        """#!/bin/bash
+'''My comment'''
+import a
+
+x = 1
+""",
+        float_to_top=True,
+        show_diff=True,
+    )
+
+
 def test_isort_shouldnt_mangle_from_multi_line_string_issue_1507():
     """isort was seen mangling lines that happened to contain the word from after
     a yield happened to be in a file. Clearly this shouldn't happen.
@@ -1121,3 +1277,524 @@ def d():
 ''',
         show_diff=True,
     )
+
+
+def test_isort_should_keep_all_as_and_non_as_imports_issue_1523():
+    """isort should keep as and non-as imports of the same path that happen to exist within the
+    same statement.
+    See: https://github.com/PyCQA/isort/issues/1523.
+    """
+    assert isort.check_code(
+        """
+from selenium.webdriver import Remote, Remote as Driver
+""",
+        show_diff=True,
+        combine_as_imports=True,
+    )
+
+
+def test_isort_shouldnt_introduce_syntax_error_issue_1539():
+    """isort should NEVER introduce syntax errors.
+    In 5.5.4 some strings that contained a line starting with from could lead to no empty paren.
+    See: https://github.com/PyCQA/isort/issues/1539.
+    """
+    assert isort.check_code(
+        '''"""Foobar
+    from {}""".format(
+    "bar",
+)
+''',
+        show_diff=True,
+    )
+    assert isort.check_code(
+        '''"""Foobar
+    import {}""".format(
+    "bar",
+)
+''',
+        show_diff=True,
+    )
+    assert (
+        isort.code(
+            '''"""Foobar
+    from {}"""
+    from a import b, a
+''',
+        )
+        == '''"""Foobar
+    from {}"""
+    from a import a, b
+'''
+    )
+    assert (
+        isort.code(
+            '''"""Foobar
+    from {}"""
+    import b
+    import a
+''',
+        )
+        == '''"""Foobar
+    from {}"""
+    import a
+    import b
+'''
+    )
+
+
+def test_isort_shouldnt_split_skip_issue_1548():
+    """Ensure isort doesn't add a spurious new line if isort: skip is combined with float to top.
+    See: https://github.com/PyCQA/isort/issues/1548.
+    """
+    assert isort.check_code(
+        """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+    prune_dependencies,
+)
+""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+    assert isort.check_code(
+        """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+    prune_dependencies,
+)
+import a
+import b
+""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+    assert isort.check_code(
+        """from tools.dependency_pruning.prune_dependencies import  # isort:skip
+import a
+import b
+""",
+        show_diff=True,
+        float_to_top=True,
+    )
+    assert isort.check_code(
+        """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+    a
+)
+import b
+""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+    assert isort.check_code(
+        """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+        )
+""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+    assert isort.check_code(
+        """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+)""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+    assert (
+        isort.code(
+            """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+)
+""",
+            profile="black",
+            float_to_top=True,
+            add_imports=["import os"],
+        )
+        == """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+)
+import os
+"""
+    )
+    assert (
+        isort.code(
+            """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+)""",
+            profile="black",
+            float_to_top=True,
+            add_imports=["import os"],
+        )
+        == """from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+)
+import os
+"""
+    )
+
+
+def test_isort_shouldnt_split_skip_issue_1556():
+    assert isort.check_code(
+        """
+from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+    prune_dependencies,
+)
+from tools.developer_pruning.prune_developers import (  # isort:skip
+    prune_developers,
+)
+""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+    assert isort.check_code(
+        """
+from tools.dependency_pruning.prune_dependencies import (  # isort:skip
+    prune_dependencies,
+)
+from tools.developer_pruning.prune_developers import x  # isort:skip
+""",
+        show_diff=True,
+        profile="black",
+        float_to_top=True,
+    )
+
+
+def test_isort_losing_imports_vertical_prefix_from_module_import_wrap_mode_issue_1542():
+    """Ensure isort doesnt lose imports when a comment is combined with an import and
+    wrap mode VERTICAL_PREFIX_FROM_MODULE_IMPORT is used.
+    See: https://github.com/PyCQA/isort/issues/1542.
+    """
+    assert (
+        isort.code(
+            """
+from xxxxxxxxxxxxxxxx import AAAAAAAAAA, BBBBBBBBBB
+from xxxxxxxxxxxxxxxx import CCCCCCCCC, DDDDDDDDD  # xxxxxxxxxxxxxxxxxx
+
+print(CCCCCCCCC)
+""",
+            multi_line_output=9,
+        )
+        == """
+from xxxxxxxxxxxxxxxx import AAAAAAAAAA, BBBBBBBBBB  # xxxxxxxxxxxxxxxxxx
+from xxxxxxxxxxxxxxxx import CCCCCCCCC, DDDDDDDDD
+
+print(CCCCCCCCC)
+"""
+    )
+
+    assert isort.check_code(
+        """
+from xxxxxxxxxxxxxxxx import AAAAAAAAAA, BBBBBBBBBB
+
+from xxxxxxxxxxxxxxxx import CCCCCCCCC, DDDDDDDDD  # xxxxxxxxxxxxxxxxxx isort: skip
+
+print(CCCCCCCCC)
+""",
+        show_diff=True,
+        multi_line_output=9,
+    )
+
+
+def test_isort_adding_second_comma_issue_1621():
+    """Ensure isort doesnt add a second comma when very long comment is present
+    See: https://github.com/PyCQA/isort/issues/1621.
+    """
+    assert isort.check_code(
+        """from .test import (
+    TestTestTestTestTestTest2 as TestTestTestTestTestTest1,  """
+        """# Some really long comment bla bla bla bla bla
+)
+""",
+        profile="black",
+        show_diff=True,
+    )
+    assert (
+        isort.code(
+            """from .test import (
+    TestTestTestTestTestTest2 as TestTestTestTestTestTest1  """
+            """# Some really long comment bla bla bla bla bla
+)
+""",
+            profile="black",
+        )
+        == """from .test import (
+    TestTestTestTestTestTest2 as TestTestTestTestTestTest1,  """
+        """# Some really long comment bla bla bla bla bla
+)
+"""
+    )
+
+
+def test_isort_shouldnt_duplicate_comments_issue_1631():
+    assert isort.check_code(
+        """
+import a  # a comment
+import a as b  # b comment
+""",
+        show_diff=True,
+    )
+    assert (
+        isort.code(
+            """
+import a  # a comment
+import a as a  # b comment
+""",
+            remove_redundant_aliases=True,
+        )
+        == """
+import a  # a comment; b comment
+"""
+    )
+
+
+def test_isort_shouldnt_add_extra_new_lines_with_import_heading_issue_1670():
+    snippet = """#!/usr/bin/python3 -ttu
+# Standard Library
+import argparse
+import datetime
+
+import attr
+import requests
+
+
+def foo() -> int:
+    print("Hello world")
+    return 0
+
+
+def spam():
+
+
+    # Standard Library
+    import collections
+    import logging
+"""
+    assert (
+        isort.code(
+            snippet,
+            import_heading_stdlib="Standard Library",
+        )
+        == snippet
+    )
+
+
+def test_isort_shouldnt_add_extra_line_float_to_top_issue_1667():
+    assert isort.check_code(
+        """
+import sys
+
+sys.path.insert(1, 'path/containing/something_else/..')
+
+import something_else  # isort:skip
+
+# Some constant
+SOME_CONSTANT = 4
+""",
+        show_diff=True,
+        float_to_top=True,
+    )
+
+
+def test_isort_shouldnt_move_noqa_comment_issue_1594():
+    assert (
+        isort.code(
+            """
+from .test import TestTestTestTestTestTest1  # noqa: F401
+from .test import TestTestTestTestTestTest2, TestTestTestTestTestTest3, """
+            """TestTestTestTestTestTest4, TestTestTestTestTestTest5  # noqa: F401
+""",
+            profile="black",
+        )
+        == """
+from .test import TestTestTestTestTestTest1  # noqa: F401
+from .test import (  # noqa: F401
+    TestTestTestTestTestTest2,
+    TestTestTestTestTestTest3,
+    TestTestTestTestTestTest4,
+    TestTestTestTestTestTest5,
+)
+"""
+    )
+
+
+def test_isort_correctly_handles_unix_vs_linux_newlines_issue_1566():
+    import_statement = (
+        "from impacket.smb3structs import (\n"
+        "SMB2_CREATE, SMB2_FLAGS_DFS_OPERATIONS, SMB2_IL_IMPERSONATION, "
+        "SMB2_OPLOCK_LEVEL_NONE, SMB2Create,"
+        "\nSMB2Create_Response, SMB2Packet)\n"
+    )
+    assert isort.code(import_statement, line_length=120) == isort.code(
+        import_statement.replace("\n", "\r\n"), line_length=120
+    ).replace("\r\n", "\n")
+
+
+def test_isort_treats_src_paths_same_as_from_config_as_cli_issue_1711(tmpdir):
+    assert isort.check_code(
+        """
+import mymodule
+import sqlalchemy
+""",
+        show_diff=True,
+    )
+
+    config_file = tmpdir.join(".isort.cfg")
+    config_file.write(
+        """
+[settings]
+src_paths=
+    api
+"""
+    )
+    api_dir = tmpdir.mkdir("api")
+    api_dir.join("mymodule.py").write("# comment")
+
+    config = isort.settings.Config(str(config_file))
+    assert isort.check_code(
+        """
+import sqlalchemy
+
+import mymodule
+""",
+        show_diff=True,
+        config=config,
+    )
+
+
+def test_isort_should_never_quietly_remove_imports_in_hanging_line_mode_issue_1741():
+    assert (
+        isort.code(
+            """
+from src import abcd, qwerty, efg, xyz  # some comment
+""",
+            line_length=50,
+            multi_line_output=2,
+        )
+        == """
+from src import abcd, efg, qwerty, xyz \\
+    # some comment
+"""
+    )
+    assert (
+        isort.code(
+            """
+from src import abcd, qwerty, efg, xyz  # some comment
+""",
+            line_length=54,
+            multi_line_output=2,
+        )
+        == """
+from src import abcd, efg, qwerty, xyz  # some comment
+"""
+    )
+    assert (
+        isort.code(
+            """
+from src import abcd, qwerty, efg, xyz  # some comment
+""",
+            line_length=53,
+            multi_line_output=2,
+        )
+        == """
+from src import abcd, efg, qwerty, xyz \\
+    # some comment
+"""
+    )
+    assert (
+        isort.code(
+            """
+from src import abcd, qwerty, efg, xyz  # some comment
+""",
+            line_length=30,
+            multi_line_output=2,
+        )
+        == """
+from src import abcd, efg, \\
+    qwerty, xyz \\
+    # some comment
+"""
+    )
+
+
+@pytest.mark.parametrize("multi_line_output", range(12))
+def test_isort_should_never_quietly_remove_imports_in_any_hangin_mode_issue_1741(
+    multi_line_output: int,
+):
+    sorted_code = isort.code(
+        """
+from src import abcd, qwerty, efg, xyz  # some comment
+""",
+        line_length=30,
+        multi_line_output=multi_line_output,
+    )
+    assert "abcd" in sorted_code
+    assert "qwerty" in sorted_code
+    assert "efg" in sorted_code
+    assert "xyz" in sorted_code
+
+
+def test_isort_should_keep_multi_noqa_with_star_issue_1744():
+    assert isort.check_code(
+        """
+from typing import *  # noqa
+from typing import IO, BinaryIO, Union  # noqa
+""",
+        show_diff=True,
+    )
+    assert isort.check_code(
+        """
+from typing import *  # noqa 1
+from typing import IO, BinaryIO, Union  # noqa 2
+""",
+        show_diff=True,
+    )
+    assert isort.check_code(
+        """
+from typing import *  # noqa
+from typing import IO, BinaryIO, Union
+""",
+        show_diff=True,
+    )
+    assert isort.check_code(
+        """
+from typing import *
+from typing import IO, BinaryIO, Union  # noqa
+""",
+        show_diff=True,
+    )
+    assert (
+        isort.code(
+            """
+from typing import *  # hi
+from typing import IO, BinaryIO, Union  # noqa
+""",
+            combine_star=True,
+        )
+        == """
+from typing import *  # noqa; hi
+"""
+    )
+    assert (
+        isort.code(
+            """
+from typing import *  # noqa
+from typing import IO, BinaryIO, Union  # noqa
+""",
+            combine_star=True,
+        )
+        == """
+from typing import *  # noqa
+"""
+    )
+
+
+def test_isort_should_keep_multiple_noqa_comments_force_single_line_mode_issue_1721():
+    assert isort.check_code(
+        """
+from some_very_long_filename_to_import_from_that_causes_a_too_long_import_row import (  # noqa: E501
+    CONSTANT_1,
+)
+from some_very_long_filename_to_import_from_that_causes_a_too_long_import_row import (  # noqa: E501
+    CONSTANT_2,
+)
+""",
+        show_diff=True,
+        profile="black",
+        force_single_line=True,
+    )
